import datetime
import json
import os

from webdriver_manager.core.config import wdm_local, get_xdist_worker_id
from webdriver_manager.core.constants import (
    DEFAULT_PROJECT_ROOT_CACHE_PATH,
    DEFAULT_USER_HOME_CACHE_PATH, ROOT_FOLDER_NAME,
)
from webdriver_manager.core.driver import Driver
from webdriver_manager.core.logger import log
from webdriver_manager.core.utils import get_date_diff, File, save_file


class DriverCache(object):
    def __init__(self, root_dir=None, valid_range=1):
        self._root_dir = DEFAULT_USER_HOME_CACHE_PATH
        is_wdm_local = wdm_local()
        xdist_worker_id = get_xdist_worker_id()
        if xdist_worker_id:
            log(f"xdist worker is: {xdist_worker_id}")
            self._root_dir = os.path.join(self._root_dir, xdist_worker_id)

        if root_dir is not None:
            self._root_dir = os.path.join(root_dir, ROOT_FOLDER_NAME, xdist_worker_id)
        if is_wdm_local:
            self._root_dir = os.path.join(DEFAULT_PROJECT_ROOT_CACHE_PATH, xdist_worker_id)

        self._drivers_root = "drivers"
        self._drivers_json_path = os.path.join(self._root_dir, "drivers.json")
        self._date_format = "%d/%m/%Y"
        self._drivers_directory = os.path.join(self._root_dir, self._drivers_root)
        self.valid_range = valid_range
        self._cache_key_driver_version = None
        self._metadata_key = None
        self._driver_binary_path = None

    def save_file_to_cache(self, driver: Driver, file: File):
        path = self.__get_path(driver)
        archive = save_file(file, path)
        files = archive.unpack(path)
        binary = self.__get_binary(files, driver.get_name())
        binary_path = os.path.join(path, binary)
        self.__save_metadata(driver, binary_path)
        log(f"Driver has been saved in cache [{path}]")
        return binary_path

    def __get_binary(self, files, driver_name):
        if len(files) == 1:
            return files[0]

        for f in files:
            if driver_name in f:
                return f

        raise Exception(f"Can't find binary for {driver_name} among {files}")

    def __save_metadata(self, driver: Driver, binary_path, date=None):
        if date is None:
            date = datetime.date.today()

        metadata = self.load_metadata_content()
        key = self.__get_metadata_key(driver)
        data = {
            key: {
                "timestamp": date.strftime(self._date_format),
                "binary_path": binary_path,
            }
        }

        metadata.update(data)
        with open(self._drivers_json_path, "w+") as outfile:
            json.dump(metadata, outfile, indent=4)

    def find_driver(self, driver: Driver):
        """Find driver by '{os_type}_{driver_name}_{driver_version}_{browser_version}'."""
        os_type = driver.get_os_type()
        driver_name = driver.get_name()
        browser_version = driver.get_browser_version_from_os()
        driver_version = self.get_cache_key_driver_version(driver)

        metadata = self.load_metadata_content()

        key = self.__get_metadata_key(driver)
        if key not in metadata:
            log(f'There is no [{os_type}] {driver_name} "{driver_version}" for browser "{browser_version}" in cache')
            return None

        driver_info = metadata[key]
        path = driver_info["binary_path"]
        if not os.path.exists(path):
            return None

        if not self.__is_valid(driver_info):
            return None

        path = driver_info["binary_path"]
        log(f"Driver [{path}] found in cache")
        return path

    def __is_valid(self, driver_info):
        dates_diff = get_date_diff(
            driver_info["timestamp"], datetime.date.today(), self._date_format
        )
        return dates_diff < self.valid_range

    def load_metadata_content(self):
        if os.path.exists(self._drivers_json_path):
            with open(self._drivers_json_path, "r") as outfile:
                return json.load(outfile)
        return {}

    def __get_metadata_key(self, driver: Driver):
        if self._metadata_key is None:
            driver_version = self.get_cache_key_driver_version(driver)
            browser_version = driver.get_browser_version_from_os()
            browser_version = browser_version if browser_version else ""
            self._metadata_key = f"{driver.get_os_type()}_{driver.get_name()}_{driver_version}_for_{browser_version}"
        return self._metadata_key

    def get_cache_key_driver_version(self, driver: Driver):
        if self._cache_key_driver_version is None:
            self._cache_key_driver_version = "latest" if driver._version in (None, "latest") else driver._version
        return self._cache_key_driver_version

    def __get_path(self, driver: Driver):
        if self._driver_binary_path is None:
            self._driver_binary_path = os.path.join(
                self._drivers_directory,
                driver.get_name(),
                driver.get_os_type(),
                driver.get_driver_version_to_download(),
            )
        return self._driver_binary_path



from webdriver_manager.core.driver import Driver
from webdriver_manager.core.logger import log


class OperaDriver(Driver):
    def __init__(
            self,
            name,
            version,
            os_type,
            url,
            latest_release_url,
            http_client,
            use_index=False,
    ):
        super(OperaDriver, self).__init__(
            name, version, os_type, url, latest_release_url, http_client, use_index,
        )

    def get_latest_release_version(self) -> str:
        resp = self._http_client.get(url=self.latest_release_url)
        return resp.text.strip()

    def get_driver_download_url(self) -> str:
        """Like https://github.com/operasoftware/operachromiumdriver/releases/download/v.2.45/operadriver_linux64.zip"""
        driver_version_to_download = self.get_driver_version_to_download()
        log(f"Getting latest opera release info for {driver_version_to_download}")
        name = "{0}_{1}.zip".format(self.get_name(), self.get_os_type())
        url = f'{self._url}/{driver_version_to_download}/{name}'
        return self._url_postprocess(url)

    @property
    def latest_release_url(self):
        return self._latest_release_url

    def get_browser_type(self):
        return "opera"



import os
from typing import Optional

from .core.download_manager import DownloadManager, WDMDownloadManager
from .core.manager import DriverManager, INDEX_SITE_ROOT, NO_INDEX_SITE
from .core.utils import ChromeType
from .drivers.chrome import ChromeDriver


class ChromeDriverManager(DriverManager):
    def __init__(
            self,
            version: Optional[str] = None,
            os_type: Optional[str] = None,
            path: Optional[str] = None,
            name: str = "chromedriver",
            url: str = f'{INDEX_SITE_ROOT}/google',
            latest_release_url: str = f'{INDEX_SITE_ROOT}/google/LATEST_RELEASE',
            chrome_type: str = ChromeType.GOOGLE,
            cache_valid_range: int = 1,
            download_manager: Optional[DownloadManager] = None,
            use_index=not NO_INDEX_SITE,
    ):
        download_manager = download_manager or WDMDownloadManager()
        driver = ChromeDriver(
            name=name,
            version=version,
            os_type=os_type,
            url=url,
            latest_release_url=latest_release_url,
            chrome_type=chrome_type,
            http_client=download_manager.http_client,
            use_index=use_index,
        )
        DriverManager.__init__(
            self, driver, path,
            cache_valid_range=cache_valid_range,
            download_manager=download_manager
        )

    def install(self) -> str:
        driver_path = self._get_driver_path(self.driver)
        os.chmod(driver_path, 0o755)
        return driver_path



import os

import pytest

from webdriver_manager.core.constants import DEFAULT_PROJECT_ROOT_CACHE_PATH
from webdriver_manager.core.download_manager import WDMDownloadManager
from webdriver_manager.core.http import WDMHttpClient
from webdriver_manager.core.utils import save_file, ChromeType
from webdriver_manager.drivers.chrome import ChromeDriver

download_manager = WDMDownloadManager()


def test_can_download_driver_as_zip_file(delete_drivers_dir):
    file = download_manager.download_file("http://chromedriver.storage.googleapis.com/2.26/chromedriver_win32.zip")
    assert file.filename == "driver.zip"
    archive = save_file(file, DEFAULT_PROJECT_ROOT_CACHE_PATH)
    assert archive.file_path == f"{DEFAULT_PROJECT_ROOT_CACHE_PATH}{os.sep}{file.filename}"
    assert archive.unpack(DEFAULT_PROJECT_ROOT_CACHE_PATH) == ["chromedriver.exe"]


def test_can_download_driver_as_tar_gz(delete_drivers_dir):
    file = download_manager.download_file(
        "https://github.com/mozilla/geckodriver/releases/download/v0.26.0/geckodriver-v0.26.0-linux32.tar.gz")
    assert file.filename == 'geckodriver-v0.26.0-linux32.tar.gz'
    archive = save_file(file, DEFAULT_PROJECT_ROOT_CACHE_PATH)
    assert archive.file_path == f"{DEFAULT_PROJECT_ROOT_CACHE_PATH}{os.sep}{file.filename}"
    assert archive.unpack(DEFAULT_PROJECT_ROOT_CACHE_PATH) == ["geckodriver"]


@pytest.mark.parametrize('version', ["2.26", None])
def test_can_download_chrome_driver(delete_drivers_dir, version):
    driver = ChromeDriver(name="chromedriver", version=version, os_type="win32",
                          url="http://chromedriver.storage.googleapis.com",
                          latest_release_url="http://chromedriver.storage.googleapis.com/LATEST_RELEASE",
                          chrome_type=ChromeType.GOOGLE, http_client=WDMHttpClient())

    file = download_manager.download_file(driver.get_driver_download_url())
    assert file.filename == "driver.zip"
    archive = save_file(file, DEFAULT_PROJECT_ROOT_CACHE_PATH)
    assert "chromedriver.exe" in archive.unpack(DEFAULT_PROJECT_ROOT_CACHE_PATH)



import requests

from webdriver_manager.core.driver import Driver
from webdriver_manager.core.logger import log


class IEDriver(Driver):
    def __init__(
            self,
            name,
            version,
            os_type,
            url,
            latest_release_url,
            http_client,
            use_index=False,
    ):
        super(IEDriver, self).__init__(
            name, version, os_type, url, latest_release_url, http_client, use_index,
        )
        self.os_type = "x64" if self._os_type == "win64" else "Win32"
        # todo: for 'browser_version' implement installed IE version detection
        #       like chrome or firefox

    def get_latest_release_version(self) -> str:
        log(f"Get LATEST driver version for Internet Explorer")
        resp = self._http_client.get(url=self.latest_release_url)
        return resp.text.strip().replace("selenium-", "")

    def _get_version_to_fulfill(self, version):
        response = requests.get(f'{self._latest_release_url}_{version}')
        if response.status_code != 404:
            response.raise_for_status()
            return response.text.strip().replace("selenium-", "")
        else:
            raise ValueError(f'Unknown version of ie webdriver - {version!r}.')

    def get_driver_version_to_download(self):
        if not self._driver_to_download_version:
            self._driver_to_download_version = self._get_version_to_fulfill(self._version) \
                if self._version not in (None, "latest") else self.get_latest_release_version()
        return self._driver_to_download_version

    def get_driver_download_url(self):
        """Like https://github.com/seleniumhq/selenium/releases/download/3.141.59/IEDriverServer_Win32_3.141.59.zip"""
        driver_version_to_download = self.get_driver_version_to_download()
        log(f"Getting latest ie release info for {driver_version_to_download}")
        filename = f"{self._name}_{self.os_type}_{driver_version_to_download}.zip"
        url = f'{self._url}/selenium-{driver_version_to_download}/{filename}'
        return self._url_postprocess(url)

    @property
    def latest_release_url(self):
        return self._latest_release_url

    def get_browser_type(self):
        return "msie"



import os

import pytest
from selenium import webdriver

from webdriver_manager.firefox import GeckoDriverManager


def test_firefox_manager_with_cache(delete_drivers_dir):
    GeckoDriverManager().install()
    binary = GeckoDriverManager().install()
    assert os.path.exists(binary)


def test_gecko_manager_with_selenium():
    driver_path = GeckoDriverManager().install()
    ff = webdriver.Firefox(executable_path=driver_path)
    ff.get("http://automation-remarks.com")
    ff.quit()


def test_driver_with_ssl_verify_disabled_can_be_downloaded(ssl_verify_enable):
    os.environ['WDM_SSL_VERIFY'] = '0'
    custom_path = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        "ssl_disabled",
    )
    driver_path = GeckoDriverManager(path=custom_path).install()
    os.environ['WDM_SSL_VERIFY'] = '1'
    assert os.path.exists(driver_path)


def test_gecko_manager_with_wrong_version():
    with pytest.raises(ValueError) as ex:
        GeckoDriverManager("0.2").install()

    assert "There is no such driver by url " in ex.value.args[0]


def test_gecko_manager_with_correct_version_and_token(delete_drivers_dir):
    driver_path = GeckoDriverManager(version="v0.11.0").install()
    assert os.path.exists(driver_path)


def test_can_download_ff_x64(delete_drivers_dir):
    driver_path = GeckoDriverManager(os_type="win64").install()
    assert os.path.exists(driver_path)


@pytest.mark.parametrize('os_type', ['win32',
                                     'win64',
                                     'linux32',
                                     'linux64',
                                     'mac64',
                                     'mac64_m1'])
def test_can_get_driver_from_cache(os_type):
    GeckoDriverManager(os_type=os_type).install()
    driver_path = GeckoDriverManager(os_type=os_type).install()
    assert os.path.exists(driver_path)



import json
import os

import pytest
from selenium import webdriver

from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.core.constants import ROOT_FOLDER_NAME


def test_chrome_manager_with_cache(delete_drivers_dir):
    ChromeDriverManager().install()
    bin = ChromeDriverManager().install()
    assert os.path.exists(bin)


def test_chrome_manager_with_specific_version(delete_drivers_dir):
    bin = ChromeDriverManager("87.0.4280.88").install()
    assert os.path.exists(bin)


def test_106_0_5249_61_chrome_version(delete_drivers_dir):
    bin = ChromeDriverManager("106.0.5249.61").install()
    assert os.path.exists(bin)


def test_chrome_manager_with_project_root_local_folder(delete_drivers_dir):
    os.environ['WDM_LOCAL'] = "1"
    bin = ChromeDriverManager("87.0.4280.88").install()
    os.environ['WDM_LOCAL'] = "0"
    assert os.path.exists(bin)


def test_driver_can_be_saved_to_custom_path():
    custom_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "custom")
    path = ChromeDriverManager(version="87.0.4280.88", path=custom_path).install()
    assert os.path.exists(path)
    assert custom_path in path


def test_chrome_manager_with_wrong_version():
    with pytest.raises(ValueError) as ex:
        ChromeDriverManager("0.2").install()
    assert "There is no such driver by url" in ex.value.args[0]


def test_chrome_manager_with_selenium():
    driver_path = ChromeDriverManager().install()
    driver = webdriver.Chrome(driver_path)
    driver.get("http://automation-remarks.com")
    driver.close()


def test_driver_with_ssl_verify_disabled_can_be_downloaded(ssl_verify_enable):
    os.environ['WDM_SSL_VERIFY'] = '0'
    custom_path = os.path.join(os.path.dirname(
        os.path.dirname(__file__)),
        "ssl_disabled",
    )
    driver_path = ChromeDriverManager(path=custom_path).install()
    os.environ['WDM_SSL_VERIFY'] = '1'
    assert os.path.exists(driver_path)


def test_chrome_manager_cached_driver_with_selenium():
    custom_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "custom-cache")
    manager = ChromeDriverManager(path=custom_path)
    driver = webdriver.Chrome(manager.install())
    driver.get("http://automation-remarks.com")

    metadata_file = os.path.join(custom_path, ROOT_FOLDER_NAME, 'drivers.json')

    with open(metadata_file) as json_file:
        data = json.load(json_file)

    for k in data.keys():
        data[k]['timestamp'] = "08/06/2019"

    with open(metadata_file, 'w') as outfile:
        json.dump(data, outfile)

    ChromeDriverManager(path=custom_path).install()


@pytest.mark.parametrize('os_type', ['win32', 'win64', 'mac64', 'mac64_m1'])
def test_can_get_chrome_for_win(os_type):
    path = ChromeDriverManager(os_type=os_type).install()
    assert os.path.exists(path)



import logging

from webdriver_manager.core.config import wdm_log_level

__logger = logging.getLogger("WDM")
__logger.addHandler(logging.NullHandler())


def log(text):
    """Emitting the log message."""
    __logger.log(wdm_log_level(), text)


def set_logger(logger):
    """
    Set the global logger.

    Parameters
    ----------
    logger : logging.Logger
        The custom logger to use.

    Returns None
    """

    # Check if the logger is a valid logger
    if not isinstance(logger, logging.Logger):
        raise ValueError("The logger must be an instance of logging.Logger")

    # Bind the logger input to the global logger
    global __logger
    __logger = logger



import glob
import os

import pytest
from selenium import webdriver
from selenium.webdriver.chrome.service import Service

from webdriver_manager.core.utils import os_type as get_os_type
from webdriver_manager.opera import OperaDriverManager


def test_opera_driver_manager_with_correct_version():
    driver_path = OperaDriverManager("v.2.45").install()
    assert os.path.exists(driver_path)


def test_driver_with_ssl_verify_disabled_can_be_downloaded(ssl_verify_enable):
    os.environ['WDM_SSL_VERIFY'] = '0'
    custom_path = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        "ssl_disabled",
    )
    driver_path = OperaDriverManager(path=custom_path).install()
    os.environ['WDM_SSL_VERIFY'] = '1'
    assert os.path.exists(driver_path)


def test_operadriver_manager_with_selenium():
    driver_path = OperaDriverManager().install()
    options = webdriver.ChromeOptions()
    options.add_experimental_option('w3c', True)

    if get_os_type() in ["win64", "win32"]:
        paths = [f for f in glob.glob(
            f"C:/Users/{os.getlogin()}/AppData/Local/Programs/Opera/**",
            recursive=True
        )]
        for path in paths:
            if os.path.isfile(path) and path.endswith("opera.exe"):
                options.binary_location = path
    elif (
            (get_os_type() in ["linux64", "linux32"])
            and not os.path.exists('/usr/bin/opera')
    ):
        options.binary_location = "/usr/bin/opera"
    elif get_os_type() in "mac64":
        options.binary_location = "/Applications/Opera.app/Contents/MacOS/Opera"
    web_service = Service(driver_path)
    web_service.start()

    opera_driver = webdriver.Remote(web_service.service_url, options=options)
    opera_driver.get("http://automation-remarks.com")
    opera_driver.quit()


def test_opera_driver_manager_with_wrong_version():
    with pytest.raises(ValueError) as ex:
        OperaDriverManager("0.2").install()

    assert "There is no such driver by url " in ex.value.args[0]


@pytest.mark.parametrize('path', ['.', None])
def test_opera_driver_manager_with_correct_version_and_token(path):
    driver_path = OperaDriverManager(version="v.2.45", path=path).install()
    assert os.path.exists(driver_path)


@pytest.mark.parametrize('os_type', ['win32',
                                     'win64',
                                     'linux64',
                                     'mac64'])
def test_can_get_driver_from_cache(os_type):
    OperaDriverManager(os_type=os_type).install()
    driver_path = OperaDriverManager(os_type=os_type).install()
    assert os.path.exists(driver_path)



import requests

from .logger import log
from .utils import get_browser_version_from_os
from .utils import os_type as _utils_os_type


class Driver(object):
    def __init__(
            self,
            name,
            version,
            os_type,
            url,
            latest_release_url,
            http_client,
            use_index=False,
    ):
        self._name = name
        self._url = url
        self._version = version
        self._os_type = os_type
        if os_type is None:
            self._os_type = _utils_os_type()
        self._latest_release_url = latest_release_url
        self._http_client = http_client
        self._browser_version = None
        self._driver_to_download_version = None
        self._use_index = bool(use_index)

    def get_name(self):
        return self._name

    def get_os_type(self):
        return self._os_type

    def get_driver_download_url(self):
        return f"{self._url}/{self.get_driver_version_to_download()}/{self._name}_{self._os_type}.zip"

    def get_driver_version_to_download(self):
        """
        Downloads version from parameter if version not None or "latest".
        Downloads latest, if version is "latest" or browser could not been determined.
        Downloads determined browser version driver in all other ways as a bonus fallback for lazy users.
        """
        if not self._driver_to_download_version:
            self._driver_to_download_version = self._version if self._version not in (
                None, "latest") else self.get_latest_release_version()
        return self._driver_to_download_version

    def get_latest_release_version(self):
        # type: () -> str
        raise NotImplementedError("Please implement this method")

    def get_browser_version_from_os(self):
        """
        Use-cases:
        - for key in metadata;
        - for printing nice logs;
        - for fallback if version was not set at all.
        Note: the fallback may have collisions in user cases when previous browser was not uninstalled properly.
        """
        if self._browser_version is None:
            self._browser_version = get_browser_version_from_os(self.get_browser_type())
        return self._browser_version

    def get_browser_type(self):
        raise NotImplementedError("Please implement this method")

    def get_binary_name(self):
        driver_name = self.get_name()
        driver_binary_name = (
            "msedgedriver" if driver_name == "edgedriver" else driver_name
        )
        driver_binary_name = (
            f"{driver_binary_name}.exe" if "win" in self.get_os_type() else driver_binary_name
        )
        return driver_binary_name

    def _url_postprocess(self, url):
        if self._use_index:
            index_url = f'{url}_index'
            log(f'Fetching resource url from index {index_url!r} ...')
            resp = requests.get(index_url)
            if resp.ok:
                return resp.text.strip()
            elif resp.status_code == 404:
                raise ValueError(f'There is no such driver by url of index - {url}.')
            else:
                resp.raise_for_status()
        else:
            resp = requests.head(url)
            if resp.ok:
                return url
            elif resp.status_code == 404:
                raise ValueError(f'There is no such driver by url {url}.')
            else:
                resp.raise_for_status()



"""tests.test_custom_logger.py"""

import logging

import pytest

from webdriver_manager.core.logger import log, set_logger, __logger

# Cache the old logger to restore it later
__old_logger = __logger


@pytest.fixture
def create_logger():
    """Create a logger."""

    # Create a Custom Debug Logger
    logger = logging.getLogger("WDM-DEBUG")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())

    return logger


def test_custom_logger(capsys, create_logger):
    """Test the custom logger."""

    # Set the custom logger
    set_logger(create_logger)

    # send a log message
    log_msg = "This is a test log message from the custom logger"
    log(log_msg)

    # Check if the log message is in the output
    captured = capsys.readouterr()
    assert log_msg in captured.err
    capsys.close()

    # Restore the old logger
    set_logger(__old_logger)



import os
import platform
from unittest import skipUnless

import pytest

from webdriver_manager.microsoft import IEDriverManager


@pytest.mark.parametrize("version", [
    "3.0",
    "3.150.0"
])
@skipUnless(platform.system() == 'Windows', 'Windows required.')
def test_ie_manager_with_different_versions(version):
    path = IEDriverManager(version).install()
    assert os.path.exists(path)


@skipUnless(platform.system() == 'Windows', 'Windows required.')
def test_driver_with_ssl_verify_disabled_can_be_downloaded(ssl_verify_enable):
    os.environ['WDM_SSL_VERIFY'] = '0'
    custom_path = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        "ssl_disabled",
    )
    driver_path = IEDriverManager(path=custom_path).install()
    os.environ['WDM_SSL_VERIFY'] = '1'
    assert os.path.exists(driver_path)


@pytest.mark.parametrize('os_type', ['win32', 'win64'])
@skipUnless(platform.system() == 'Windows', 'Windows required.')
def test_can_download_ie_driver_x64(os_type):
    path = IEDriverManager(os_type=os_type).install()
    assert os.path.exists(path)


@pytest.mark.parametrize('os_type', ['win32', 'win64'])
@skipUnless(platform.system() == 'Windows', 'Windows required.')
def test_can_get_ie_driver_from_cache(os_type):
    IEDriverManager(os_type=os_type).install()
    driver_path = IEDriverManager(os_type=os_type).install()
    assert os.path.exists(driver_path)



import os

from .download_manager import WDMDownloadManager
from .driver import Driver
from .driver_cache import DriverCache
from .logger import log

INDEX_SITE_ROOT = os.environ.get(
    'INDEX_SITE_ROOT',
    'https://gitee.com/hansbug/browser_drivers_mirror_index/raw/master',
)
NO_INDEX_SITE = bool(os.environ.get('NO_INDEX_SITE', '').strip())


class DriverManager(object):
    def __init__(
            self,
            driver: Driver,
            root_dir=None,
            cache_valid_range=1,
            download_manager=None
    ):
        self.driver = driver
        self.driver_cache = DriverCache(root_dir, cache_valid_range)
        self._download_manager = download_manager
        if download_manager is None:
            self._download_manager = WDMDownloadManager()
        log("====== WebDriver manager ======")

    @property
    def http_client(self):
        return self._download_manager.http_client

    def install(self) -> str:
        raise NotImplementedError  # pragma: no cover

    def _get_driver_path(self, driver):
        binary_path = self.driver_cache.find_driver(driver)
        if binary_path:
            return binary_path

        url = driver.get_driver_download_url()
        file = self._download_manager.download_file(url)
        binary_path = self.driver_cache.save_file_to_cache(driver, file)
        return binary_path

    @property
    def latest_version(self):
        return self.driver.get_latest_release_version()

    @property
    def version_to_download(self):
        return self.driver.get_driver_version_to_download()

    @property
    def driver_url(self):
        return self.driver.get_driver_download_url()

    @property
    def browser_version(self):
        return self.driver.get_browser_version_from_os()

    @property
    def driver_executable(self):
        return self.install()



import logging
import os

import pytest

from webdriver_manager.chrome import ChromeDriverManager


@pytest.fixture
def log():
    os.environ['WDM_LOG'] = str(logging.NOTSET)
    yield
    os.environ['WDM_LOG'] = str(logging.INFO)


def test_chrome_manager_with_specific_version(log):
    bin = ChromeDriverManager("87.0.4280.88").install()
    assert os.path.exists(bin)



import datetime
import os
import platform
import re
import subprocess
import sys

from tqdm import tqdm

from webdriver_manager.core.archive import Archive


class File(object):
    def __init__(self, stream):
        self.content = stream.content
        self.__stream = stream
        self.__temp_name = "driver"

    @property
    def filename(self) -> str:
        try:
            filename = re.findall(
                r'filename=\s*[\"\']?(?P<filename>[^\s\"\';]+)[\"\'\s]?',
                self.__stream.headers["content-disposition"],
            )[0]
        except KeyError:
            filename = f"{self.__temp_name}.zip"
        except IndexError:
            filename = f"{self.__temp_name}.exe"

        if '"' in filename:
            filename = filename.replace('"', "")

        return filename


def save_file(file: File, directory: str):
    os.makedirs(directory, exist_ok=True)

    archive_path = f"{directory}{os.sep}{file.filename}"
    with open(archive_path, "wb") as code:
        code.write(file.content)
    if not os.path.exists(archive_path):
        raise FileExistsError(f"No file has been saved on such path {archive_path}")
    return Archive(archive_path, os_type=os_name())


class OSType(object):
    LINUX = "linux"
    MAC = "mac"
    WIN = "win"


class ChromeType(object):
    GOOGLE = "google-chrome"
    CHROMIUM = "chromium"
    BRAVE = "brave-browser"
    MSEDGE = "edge"


PATTERN = {
    ChromeType.CHROMIUM: r"\d+\.\d+\.\d+",
    ChromeType.GOOGLE: r"\d+\.\d+\.\d+",
    ChromeType.MSEDGE: r"\d+\.\d+\.\d+",
    "brave-browser": r"(\d+)",
    "firefox": r"(\d+.\d+)",
}


def os_name():
    pl = sys.platform
    if pl == "linux" or pl == "linux2":
        return OSType.LINUX
    elif pl == "darwin":
        return OSType.MAC
    elif pl == "win32":
        return OSType.WIN


def os_architecture():
    if platform.machine().endswith("64"):
        return 64
    else:
        return 32


def os_type():
    return f"{os_name()}{os_architecture()}"


def is_arch(os_sys_type):
    if '_m1' in os_sys_type:
        return True
    return platform.processor() != 'i386'


def is_mac_os(os_sys_type):
    return OSType.MAC in os_sys_type


def get_date_diff(date1, date2, date_format):
    a = datetime.datetime.strptime(date1, date_format)
    b = datetime.datetime.strptime(
        str(date2.strftime(date_format)), date_format)

    return (b - a).days


def linux_browser_apps_to_cmd(*apps: str) -> str:
    """Create 'browser --version' command from browser app names.

    Result command example:
        chromium --version || chromium-browser --version
    """
    ignore_errors_cmd_part = " 2>/dev/null" if os.getenv("WDM_LOG_LEVEL") == "0" else ""
    return " || ".join(f"{i} --version{ignore_errors_cmd_part}" for i in apps)


def windows_browser_apps_to_cmd(*apps: str) -> str:
    """Create analogue of browser --version command for windows."""
    powershell = determine_powershell()

    first_hit_template = """$tmp = {expression}; if ($tmp) {{echo $tmp; Exit;}};"""
    script = "$ErrorActionPreference='silentlycontinue'; " + " ".join(
        first_hit_template.format(expression=e) for e in apps
    )

    return f'{powershell} -NoProfile "{script}"'


def get_browser_version_from_os(browser_type=None):
    """Return installed browser version."""
    cmd_mapping = {
        ChromeType.GOOGLE: {
            OSType.LINUX: linux_browser_apps_to_cmd(
                "google-chrome",
                "google-chrome-stable",
                "google-chrome-beta",
                "google-chrome-dev",
            ),
            OSType.MAC: r"/Applications/Google\ Chrome.app/Contents/MacOS/Google\ Chrome --version",
            OSType.WIN: windows_browser_apps_to_cmd(
                r'(Get-Item -Path "$env:PROGRAMFILES\Google\Chrome\Application\chrome.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\Google\Chrome\Application\chrome.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:LOCALAPPDATA\Google\Chrome\Application\chrome.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\Google\Chrome\BLBeacon").version',
                r'(Get-ItemProperty -Path Registry::"HKLM\SOFTWARE\Wow6432Node\Microsoft\Windows\CurrentVersion\Uninstall\Google Chrome").version',
            ),
        },
        ChromeType.CHROMIUM: {
            OSType.LINUX: linux_browser_apps_to_cmd("chromium", "chromium-browser"),
            OSType.MAC: r"/Applications/Chromium.app/Contents/MacOS/Chromium --version",
            OSType.WIN: windows_browser_apps_to_cmd(
                r'(Get-Item -Path "$env:PROGRAMFILES\Chromium\Application\chrome.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\Chromium\Application\chrome.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:LOCALAPPDATA\Chromium\Application\chrome.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\Chromium\BLBeacon").version',
                r'(Get-ItemProperty -Path Registry::"HKLM\SOFTWARE\Wow6432Node\Microsoft\Windows\CurrentVersion\Uninstall\Chromium").version',
            ),
        },
        ChromeType.BRAVE: {
            OSType.LINUX: linux_browser_apps_to_cmd(
                "brave-browser", "brave-browser-beta", "brave-browser-nightly"
            ),
            OSType.MAC: r"/Applications/Brave\ Browser.app/Contents/MacOS/Brave\ Browser --version",
            OSType.WIN: windows_browser_apps_to_cmd(
                r'(Get-Item -Path "$env:PROGRAMFILES\BraveSoftware\Brave-Browser\Application\brave.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\BraveSoftware\Brave-Browser\Application\brave.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:LOCALAPPDATA\BraveSoftware\Brave-Browser\Application\brave.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\BraveSoftware\Brave-Browser\BLBeacon").version',
                r'(Get-ItemProperty -Path Registry::"HKLM\SOFTWARE\Wow6432Node\Microsoft\Windows\CurrentVersion\Uninstall\BraveSoftware Brave-Browser").version',
            ),
        },
        ChromeType.MSEDGE: {
            OSType.LINUX: linux_browser_apps_to_cmd(
                "microsoft-edge",
                "microsoft-edge-stable",
                "microsoft-edge-beta",
                "microsoft-edge-dev",
            ),
            OSType.MAC: r"/Applications/Microsoft\ Edge.app/Contents/MacOS/Microsoft\ Edge --version",
            OSType.WIN: windows_browser_apps_to_cmd(
                # stable edge
                r'(Get-Item -Path "$env:PROGRAMFILES\Microsoft\Edge\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\Microsoft\Edge\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\Microsoft\Edge\BLBeacon").version',
                r'(Get-ItemProperty -Path Registry::"HKLM\SOFTWARE\Microsoft\EdgeUpdate\Clients\{56EB18F8-8008-4CBD-B6D2-8C97FE7E9062}").pv',
                # beta edge
                r'(Get-Item -Path "$env:LOCALAPPDATA\Microsoft\Edge Beta\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES\Microsoft\Edge Beta\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\Microsoft\Edge Beta\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\Microsoft\Edge Beta\BLBeacon").version',
                # dev edge
                r'(Get-Item -Path "$env:LOCALAPPDATA\Microsoft\Edge Dev\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES\Microsoft\Edge Dev\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\Microsoft\Edge Dev\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\Microsoft\Edge Dev\BLBeacon").version',
                # canary edge
                r'(Get-Item -Path "$env:LOCALAPPDATA\Microsoft\Edge SxS\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\Microsoft\Edge SxS\BLBeacon").version',
                # highest edge
                r"(Get-Item (Get-ItemProperty 'HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\msedge.exe').'(Default)').VersionInfo.ProductVersion",
                r"[System.Diagnostics.FileVersionInfo]::GetVersionInfo((Get-ItemProperty 'HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\msedge.exe').'(Default)').ProductVersion",
                r"Get-AppxPackage -Name *MicrosoftEdge.* | Foreach Version",
                r'(Get-ItemProperty -Path Registry::"HKLM\SOFTWARE\Wow6432Node\Microsoft\Windows\CurrentVersion\Uninstall\Microsoft Edge").version',
            ),
        },
        "firefox": {
            OSType.LINUX: linux_browser_apps_to_cmd("firefox"),
            OSType.MAC: r"/Applications/Firefox.app/Contents/MacOS/firefox --version",
            OSType.WIN: windows_browser_apps_to_cmd(
                r'(Get-Item -Path "$env:PROGRAMFILES\Mozilla Firefox\firefox.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\Mozilla Firefox\firefox.exe").VersionInfo.FileVersion',
                r"(Get-Item (Get-ItemProperty 'HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\firefox.exe').'(Default)').VersionInfo.ProductVersion",
                r'(Get-ItemProperty -Path Registry::"HKLM\SOFTWARE\Mozilla\Mozilla Firefox").CurrentVersion',
            ),
        },
    }

    try:
        cmd_mapping = cmd_mapping[browser_type][os_name()]
        pattern = PATTERN[browser_type]
        version = read_version_from_cmd(cmd_mapping, pattern)
        return version
    except Exception:
        return None


def read_version_from_cmd(cmd, pattern):
    with subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            stdin=subprocess.DEVNULL,
            shell=True,
    ) as stream:
        stdout = stream.communicate()[0].decode()
        version = re.search(pattern, stdout)
        version = version.group(0) if version else None
    return version


def determine_powershell():
    """Returns "True" if runs in Powershell and "False" if another console."""
    cmd = "(dir 2>&1 *`|echo CMD);&<# rem #>echo powershell"
    with subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            stdin=subprocess.DEVNULL,
            shell=True,
    ) as stream:
        stdout = stream.communicate()[0].decode()
    return "" if stdout == "powershell" else "powershell"


def show_download_progress(response, _bytes_threshold=100):
    """ Opens up a response's content in chunks to show a progress bar with tqdm.
        Resets response._content when done so that response can be consumed again as normal. """
    total = int(response.headers.get("Content-Length", 0))
    if total > _bytes_threshold:
        content = bytearray()
        progress_bar = tqdm(desc="[WDM] - Downloading", total=total, unit_scale=True, unit_divisor=1024, unit="B")
        for chunk in response.iter_content(chunk_size=8192):
            if chunk:  # Filter out keep-alive new chunks
                progress_bar.update(len(chunk))
                content.extend(chunk)
        response._content = content  # To allow content to be "consumed" again



from typing import Type

from .chrome import ChromeDriverManager
from .core.manager import DriverManager
from .firefox import GeckoDriverManager
from .microsoft import IEDriverManager, EdgeChromiumDriverManager
from .opera import OperaDriverManager

_BROWSER_ALIAS = {
    'msie': 'ie',
    'msedge': 'edge',
    'google': 'chrome',
    'gecko': 'firefox',
}
_BROWSER_MANAGERS = {
    'ie': IEDriverManager,
    'edge': EdgeChromiumDriverManager,
    'chrome': ChromeDriverManager,
    'firefox': GeckoDriverManager,
    'opera': OperaDriverManager,
}


def _parse_browser_name(name: str):
    original_name = name
    name = name.lower().strip()
    name = _BROWSER_ALIAS.get(name, name)
    if name in _BROWSER_MANAGERS:
        return name
    else:
        raise ValueError(f'Unknown browser name - {original_name!r}.')


def get_browser_manager_class(name) -> Type[DriverManager]:
    return _BROWSER_MANAGERS[_parse_browser_name(name)]


def get_browser_manager(browser_name: str, *args, **kwargs):
    return get_browser_manager_class(browser_name)(*args, **kwargs)



import os
from typing import Optional

from .core.download_manager import DownloadManager, WDMDownloadManager
from .core.manager import DriverManager, INDEX_SITE_ROOT, NO_INDEX_SITE
from .drivers.opera import OperaDriver


class OperaDriverManager(DriverManager):
    def __init__(
            self,
            version: Optional[str] = None,
            os_type: Optional[str] = None,
            path: Optional[str] = None,
            name: str = "operadriver",
            url: str = f"{INDEX_SITE_ROOT}/opera",
            latest_release_url: str = f"{INDEX_SITE_ROOT}/opera/LATEST_RELEASE",
            cache_valid_range: int = 1,
            download_manager: Optional[DownloadManager] = None,
            use_index=not NO_INDEX_SITE,
    ):
        download_manager = download_manager or WDMDownloadManager()
        driver = OperaDriver(
            name=name,
            version=version,
            os_type=os_type,
            url=url,
            latest_release_url=latest_release_url,
            http_client=download_manager.http_client,
            use_index=use_index,
        )
        DriverManager.__init__(
            self, driver, path, cache_valid_range,
            download_manager=download_manager
        )

    def install(self) -> str:
        driver_path = self._get_driver_path(self.driver)
        if not os.path.isfile(driver_path):
            for name in os.listdir(driver_path):
                if "sha512_sum" in name:
                    os.remove(os.path.join(driver_path, name))
                    break
        driver_path = os.path.join(driver_path, os.listdir(driver_path)[0])
        os.chmod(driver_path, 0o755)
        return driver_path
